# destine_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import xarray as xr
from polytope.api import Client
import glob
import shutil
from cdo import Cdo
from importlib.resources import files


def load_destine(model_global, variable, experiment, start, end,
                 output_path):
    """
    Download hourly Destination Earth DT data via the Polytope web service
    hosted on the LUMI Databridge.

        Input:
            model_global    str     name of the global model
            variable        str     variable name (e.g. 2 metre temperature)
            experiment      str     climate change scenario
            start           str     start (e.g. 20200901)
            end             str     end (e.g. 20200930)
            output_path     str     path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the DestinE
            daily mean data calculated from downloaded hourly data.
    """
    # Define filename
    file = f"{variable}_{model_global}_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded data successfully.")
        return output_file

    else:
        # Define temporary directory
        output_path_temp = os.path.join(output_path, "tmp")

        # Ensure the temporary directory exists
        os.makedirs(output_path_temp, exist_ok=True)

        # Define filename for temporal output
        temp_file = os.path.join(output_path_temp, file)

        # Define range of dates to be downloaded
        dates = xr.cftime_range(
            start=start,
            end=end,
            freq='1D',
            inclusive='both', 
            calendar='proleptic_gregorian'
        )

        for date in dates:
            date = date.strftime('%Y%m%d')

            fn = os.path.join(output_path_temp, f'{model_global}-climate-dt_{variable}_high_hourly_{date}.grib')
            if os.path.isfile(fn):
                continue

            # Initialize client
            client = Client(
                address="polytope.lumi.apps.dte.destination-earth.eu",
            )

            # Optionally revoke previous requests
            client.revoke("all")

            # Check experiment/activity consistency for scenario simulation
            if experiment == 'SSP3-7.0':
                activity = 'ScenarioMIP'

            # Check experiment/activity consistency for historical simulation
            if experiment == 'hist':
                activity = 'CMIP6'

            # Check variable consistency for temperature
            if variable == 'tas':
                variable_request = '167'

            # Check variable consistency for precipitation
            if variable == 'pr':
                variable_request = '260048'

            request = {
                "activity": activity,
                "class": "d1",
                "dataset": "climate-dt",
                "date": date,
                "experiment": experiment,
                "expver": "0001",
                "generation": "1",
                "levtype": "sfc",
                "model": model_global,
                "param": variable_request,
                "realization": "1",
                "resolution": "high",
                "stream": "clte",
                "time": "0000/0100/0200/0300/0400/0500/0600/0700/0800/0900/1000/1100/1200/1300/1400/1500/1600/1700/1800/1900/2000/2100/2200/2300",
                "type": "fc",
            }

            try:
                retrieve = client.retrieve(
                    "destination-earth",
                    request, 
                    output_file=fn, 
                    pointer=False)
                print(f'{date=}...DONE')

            except Exception:
                print(f'{date=}...FAIL')
                continue

        # Glob all GRIB files in the directory
        grib_files = sorted(glob.glob(os.path.join(output_path_temp, "*.grib")))

        for grib_file in grib_files:
            # Open the GRIB file
            dataset = xr.open_dataset(grib_file, engine="cfgrib")

            if variable == 'tas':
                # Rename the variable
                dataset = dataset.rename({'t2m': variable})

                # Calculate the daily mean
                daily_mean = dataset.resample(valid_time="1D").mean(dim="time")

                # Rename the time coordinate
                daily_mean = daily_mean.rename({'valid_time': 'time'})

                # Convert from Kelvin to Celsius
                daily_mean[variable] = daily_mean[variable] - 273.15

                # Generate an output file name
                base_name = os.path.basename(grib_file).replace(".grib", "_daily_mean.nc")
                daily_file = os.path.join(output_path_temp, base_name)

                # Save the daily mean to a NetCDF file
                daily_mean.to_netcdf(daily_file)

            if variable == 'pr':
                # Rename the variable
                dataset = dataset.rename({'tprate': variable})

                # Calculate the daily mean
                daily_mean = dataset.resample(valid_time="1D").mean(dim="time")

                # Rename the time coordinate
                daily_mean = daily_mean.rename({'valid_time': 'time'})

                # Convert from kg/m² second into kg/m² day
                daily_mean[variable] = daily_mean[variable]*86400

                # Generate an output file name
                base_name = os.path.basename(grib_file).replace(".grib", "_daily_mean.nc")
                daily_file = os.path.join(output_path_temp, base_name)

                # Save the daily mean to a NetCDF file
                daily_mean.to_netcdf(daily_file)

        # Use CDO to merge netcdf files
        daily_files = sorted(glob.glob(os.path.join(output_path_temp, "*_daily_mean.nc")))

        if daily_files:
            # Convert the list of files into a space-separated string
            file_list = " ".join(daily_files)

            # If cdo is not in the path, add it manually
            conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
            os.environ['PATH'] += f':{conda_bin}'

            # Initialize the Cdo object
            cdo = Cdo()

            # Perform merging of files on the time dimension
            cdo.mergetime(input=file_list, output=temp_file)

            # Dynamically get the file path to healpix_grid.txt inside the package
            healpix_grid_path = files('climxtract').joinpath('healpix_grid.txt')

            # Set grid from unstructured (default Destination Earth) to healpix for conservative regridding
            cdo.setgrid(str(healpix_grid_path), input=temp_file, output=output_file)

        # Path to the directory you want to remove
        shutil.rmtree(output_path_temp)

        return output_file


# spartacus_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import shutil
import wget
import xarray as xr
from cdo import Cdo
from importlib.resources import files


def download_yearly_files(dataset, year, output_path):
    """
    Download SPARTACUS v2.1 data on a yearly basis from the
    Geosphere Austria Datahub.

        Input:
            dataset    str      type of the dataset (e.g tn)
            year       int      year to be downloaded (e.g 1971)

        Output:
            Returns the temporary output path, where the yearly downloaded
            files are saved.
    """
    # Putting together the download link (url)
    base_url = (
        "https://public.hub.geosphere.at/datahub/resources/"
        "spartacus-v2-1d-1km/filelisting"
    )

    url_directory = f"{base_url}/{dataset.upper()}/"
    file = f"SPARTACUS2-DAILY_{dataset.upper()}_{year}.nc"

    # Construct the full URL
    url = f"{url_directory}{file}"

    # Define temporary directory
    output_path_temp = os.path.join(output_path, "tmp")

    # Ensure the temporary directory exsists
    os.makedirs(output_path_temp, exist_ok=True)

    # Define filname for temporal output
    temp_file = os.path.join(output_path_temp, file)

    # Download the file using the wget package
    wget.download(url, out=temp_file)

    return temp_file


def load_spartacus(variable, start, end, output_path):
    """
    Function that calls download_netcdf_files to download SPARTACUS data.
    The individual downloaded years are merged into one file.
    The temperature is calculated as the mean of daily maximum and
    minimum temperature.

        Input:
            variable          str      variable name (e.g. tas)
            start             str      start year (e.g. 1961)
            end               str      end year (e.g. 2024)
            output_path       str      path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the SPARTACUS data.
    """
    # Define filename
    tmp_file = (
        f"{variable}_spartacusv2.1_"
        f"{start}-{end}_tmp.nc"
    )

    file = (
        f"{variable}_spartacusv2.1_"
        f"{start}-{end}.nc"
    )

    tmp_file = os.path.join(output_path, tmp_file)
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded data successfully.")
        return output_file

    else:

        try:
            # Initialize list to hold data arrays
            data_list = []

            # Loop through years and download data
            for year in range(int(start), int(end) + 1):
                if variable == 'tas':
                    # Download tn and tx datasets
                    tasmin_path = download_yearly_files('tn', year, output_path)
                    tasmax_path = download_yearly_files('tx', year, output_path)

                    # Load the datasets
                    tasmin = xr.open_dataset(tasmin_path)
                    tasmax = xr.open_dataset(tasmax_path)

                    # Append the data arrays to the list
                    data_list.append((tasmin['TN'] + tasmax['TX']) / 2)

                elif variable == 'pr':
                    # Download rr dataset
                    rr_path = download_yearly_files('rr', year, output_path)

                    # Load the dataset
                    rr = xr.open_dataset(rr_path)

                    # Appaned the data arrays to the list
                    data_list.append(rr['RR'])

            # Concatenate all years together along the time dimension
            combined_data = xr.concat(data_list, dim='time')
            combined_data.name = variable

            # Save the resulting dataset to a single NetCDF file
            combined_data.to_netcdf(tmp_file)
            print("Downloaded and processed data successfully.")

            # If cdo is not in the path, add it manually
            conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
            os.environ['PATH'] += f':{conda_bin}'

            # Initialize the Cdo object
            cdo = Cdo()

            # Dynamically get the file path to spartacus_grid.txt inside the package
            spartacus_grid_path = files('climxtract').joinpath('spartacus_grid.txt')

            # Set grid from unstructured (default Destination Earth) to healpix for conservative regridding
            cdo.setgrid(str(spartacus_grid_path), input=tmp_file, output=output_file)

        except Exception as e:
            print("Failed to download and process data.")
            print(f"\nAn error occured: {e}")

        # Path to the directory you want to remove
        shutil.rmtree(os.path.join(output_path, "tmp"))

        # remove tmp_file with old grid
        os.remove(tmp_file)

        return output_file


# eobs_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import wget
import shutil
from cdo import Cdo


def load_eobs(variable, start, end, output_path):
    """
    Download daily E-OBS data from European Climate Assessment & Dataset.
    The complete dataset from 1950 onwards is downloaded and saved.

        Input:
            variable       str          variable name (e.g. 2m_temperature)
            start          str          start year (1950)
            end            str          end year (2024)
            output_path    str          path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the ERA5 daily mean data.
    """
    # Define filename
    file = f"{variable}_e-obs30.e_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded data successfully.")
        return output_file

    else:
        # Check variable consistency for temperature
        if variable == 'tas':
            variable_long = 'tg'

        # Check variable consistency for precipitation
        if variable == 'pr':
            variable_long = 'rr'

        # Define temporary directory
        output_path_temp = os.path.join(output_path, "tmp")

        # Ensure the temporary directory exists
        os.makedirs(output_path_temp, exist_ok=True)

        # Define filename for temporal output
        temp_file = os.path.join(output_path_temp, file)

        try:
            # Putting together the download link (url)
            base_url = "https://knmi-ecad-assets-prd.s3.amazonaws.com/ensembles/data/"
            url_directory = base_url + "Grid_0.1deg_reg_ensemble/"
            filename = str(variable_long) + "_" + "ens_mean_0.1deg_reg_v30.0e.nc"
            url = url_directory + filename

            # Download the file using the wget package
            wget.download(url, out=temp_file)
            print("Downloaded data successfully.")

        except Exception as e:
            print("Failed to download data.")
            print(f"\nAn error occured: {e}")

        # If cdo is not in the path, add it manually
        conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
        os.environ['PATH'] += f':{conda_bin}'

        # Initialize the Cdo object
        cdo = Cdo()

        if variable == 'tas':
            # Rename the variable
            cdo.chname("tg", "tas", input=temp_file, output=output_file)

        if variable == 'pr':
            # Rename the variable
            cdo.chname("rr", "pr", input=temp_file, output=output_file)

        # Path to the directory you want to remove
        shutil.rmtree(output_path_temp)

        return output_file


# regrid.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
from cdo import Cdo
from importlib.resources import files


def remapnn(target_file, source_file, output_path_regrid):
    """
    Remapping any grid to lambert_conformal_conic projection using nearest-neighbour interpolation.
    Target file can be either a netcdf-file or a grid description file (.txt).

        Input:
            target_file            str   path to the target file
            source_file            str   path to the source file
            output_path_regrid     str   base path where the output file is stored

        Output:
            Returns the remapped netcdf-file as well as the output path for the cutting function.
    """
    # If cdo is not in the path, add it manually
    conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
    os.environ['PATH'] += f':{conda_bin}'

    # Extract the filename from the source file
    filename = source_file.split('/')[-1]

    # Modify the filename to include '_remapnn' before the .nc extension
    file = filename.replace(".nc", "_remapnn.nc")
    output_file = f"{output_path_regrid}/{file}"

    # Check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        return output_file

    else:
        # Initialize the Cdo object
        cdo = Cdo()

        # Perform the regridding using the remapnn method
        cdo.remapnn(target_file, input=source_file, output=output_file)
        return output_file


def remapbil(target_file, source_file, output_path_regrid):
    """
    Remapping any grid to lambert_conformal_conic projection using bilinear interpolation.
    Target file can be either a netcdf-file or a grid description file (.txt).

        Input:
            target_file              str     path to the target file
            source_file              str     path to the source file
            output_path_regrid       str     base path where the output file is saved

        Output:
            Returns the remapped netcdf-file as well as the output path for the cutting function.
    """
    # If cdo is not in the path, add it manually
    conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
    os.environ['PATH'] += f':{conda_bin}'

    # Extract the filename from the source file
    filename = source_file.split('/')[-1]

    # Modify the filename to include '_remapbil' before the .nc extension
    file = filename.replace(".nc", "_remapbil.nc")
    output_file = f"{output_path_regrid}/{file}"

    # Check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        return output_file

    else:
        # Initialize the Cdo object
        cdo = Cdo()

        # Perform the regridding using the remapbil method
        cdo.remapbil(target_file, input=source_file, output=output_file)
        return output_file


def remapcon(target_file, source_file, output_path_regrid):
    """
    Remapping any grid to lambert_conformal_conic projection using conservative interpolation.
    Target file can be either a netcdf-file or a grid description file (.txt).

        Input:
            target_file              str     path to the target file
            source_file              str     path to the source file
            output_path_regrid       str     base path where the output file is saved

        Output:
            Returns the remapped netcdf-file as well as the output path for the cutting function.
    """
    # If cdo is not in the path, add it manually
    conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
    os.environ['PATH'] += f':{conda_bin}'

    # Extract the filename from the source file
    filename = source_file.split('/')[-1]

    # Modify the filename to include '_remapcon' before the .nc extension
    file = filename.replace(".nc", "_remapcon.nc")
    output_file = f"{output_path_regrid}/{file}"

    # Check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        return output_file

    else:
        # Initialize the Cdo object
        cdo = Cdo()

        # Dynamically get the file path to oeks15_grid.txt inside the package
        oeks15_grid_path = files('climxtract').joinpath('oeks15_grid.txt')

        # Perform the regridding using the remapcon method
        cdo.remapcon(str(oeks15_grid_path), input=source_file, output=output_file, options="--force")
        return output_file


def remapdis(target_file, source_file, output_path_regrid):
    """
    Remapping any grid to lambert_conformal_conic projection using distance weighted interpolation.
    Target file can be either a netcdf-file or a grid description file (.txt).

        Input:
            target_file              str     path to the target file
            source_file              str     path to the source file
            output_path_regrid       str     base path where the output file is saved

        Output:
            Returns the remapped netcdf-file as well as the output path for the cutting function.
    """
    # If cdo is not in the path, add it manually
    conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
    os.environ['PATH'] += f':{conda_bin}'

    # Extract the filename from the source file
    filename = source_file.split('/')[-1]

    # Modify the filename to include '_remapdis' before the .nc extension
    file = filename.replace(".nc", "_remapdis.nc")
    output_file = f"{output_path_regrid}/{file}"

    # check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        return output_file
    else:
        # Initialize the Cdo object
        cdo = Cdo()

        # Perform the regridding using the remapdis method
        cdo.remapdis(target_file, input=source_file, output=output_file)
        return output_file


# era5_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import cdsapi
import zipfile
import os
import xarray as xr
import numpy as np
from datetime import datetime, timedelta


def load_era5(model_global, variable, start, end, output_path):
    """
    Download hourly ERA5 data from the Copernicus Climate Data Store using the
    CDS API and convert hourly data into daily means.

        Input:
            model_global          str          name of the dataset
            variable_long         str          variable name
            start                 str          start date (e.g. 20200901)
            end                   str          end date (e.g. 20200930)
            output_path           str          path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the ERA5 daily mean
            data calculated from downloaded hourly data.
    """
    # Define filename
    file = f"{variable}_{model_global}_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded data successfully.")
        return output_file

    else:
        # Check variable consistency for temperature
        if variable == 'tas':
            variable_long = '2m_temperature'

        # Check variable consistency for precipitation
        if variable == 'pr':
            variable_long = 'total_precipitation'

        try:
            # Initialize the API client
            client = cdsapi.Client()

            # Convert strings to datetime objects
            start_date = datetime.strptime(start, "%Y%m%d")
            end_date = datetime.strptime(end, "%Y%m%d")

            # Generate list of years, months, and days
            date_range = [start_date + timedelta(days=x) for x in range((end_date - start_date).days + 1)]

            years = sorted(list(set(date.strftime("%Y") for date in date_range)))
            months = sorted(list(set(date.strftime("%m") for date in date_range)))
            days = sorted(list(set(date.strftime("%d") for date in date_range)))

            dataset = model_global
            request = {
                "product_type": ["reanalysis"],
                "variable": variable_long,
                "year": years,
                "month": months,
                "day": days,
                "time": [
                    "00:00", "01:00", "02:00",
                    "03:00", "04:00", "05:00",
                    "06:00", "07:00", "08:00",
                    "09:00", "10:00", "11:00",
                    "12:00", "13:00", "14:00",
                    "15:00", "16:00", "17:00",
                    "18:00", "19:00", "20:00",
                    "21:00", "22:00", "23:00"
                ],
                "area": [80, -50, 20, 75],
                "data_format": "netcdf",
                "download_format": "zip"
            }

            target = os.path.join(output_path, 'download.zip')

            client.retrieve(dataset, request, target)
            print("Downloaded data successfully.")

        except Exception as e:
            print("Failed to download data.")
            print(f"\nAn error occured: {e}")

        # Unzipping the file and saving the contents to the specified directory
        with zipfile.ZipFile(target, 'r') as zip_ref:
            # Extract all files
            zip_ref.extractall(output_path)

            # List the files inside the zip
            extracted_files = zip_ref.namelist()

        # Get the full path of each extracted file
        filename = [os.path.join(output_path, file) for file in extracted_files]

        # Take list and convert into a string
        filename = ' '.join(filename)

        # Remove downloaded zip file
        os.remove(target)

        # Open previously downloaded hourly dataset
        dataset = xr.open_dataset(filename)

        if variable == 'tas':

            # Rename the variable
            dataset = dataset.rename({'t2m': variable})

            # Calculate daily means based on hourly data
            daily_mean = dataset.resample(valid_time='1D').mean(dim='valid_time')

            # Rename the time coordinate
            daily_mean = daily_mean.rename({'valid_time': 'time'})

            # Convert from kelvin to celsius
            daily_mean[variable] = daily_mean[variable] - 273.15

            # write dataset contents to a netCDF file
            daily_mean.to_netcdf(output_file)

        # ERA5-Land hourly precip data to total precip for a day
        if variable == 'pr' and model_global == 'reanalysis-era5-land':

            # Rename the variable
            dataset = dataset.rename({'tp': variable})

            # Get all 00UTC time steps, because it contains the accumulated precip
            daily_sum = dataset.sel(valid_time=dataset.valid_time.dt.hour == 0)

            # Assign the precipitation to the previous day explicitly:
            daily_sum = daily_sum.assign_coords(
                valid_time=(daily_sum.valid_time - np.timedelta64(1, 'D'))
            )

            # Rename the time coordinate
            daily_sum = daily_sum.rename({'valid_time': 'time'})

            # Convert from meters to millimeters
            daily_sum[variable] = daily_sum[variable]*1000

            # Write dataset contents to a netCDF file
            daily_sum.to_netcdf(output_file)

        # ERA5 hourly precip data to total precip for a day
        if variable == 'pr' and model_global == 'reanalysis-era5-single-levels':

            # Rename the variable
            dataset = dataset.rename({'tp': variable})

            # Convert from m to mm
            dataset[variable] = dataset[variable]*1000

            # Round times to the nearest hour to avoid datetime issues
            dataset['valid_time'] = dataset['valid_time'].dt.round('1h')

            # Prepare output list
            daily_pr_list = []

            # Get date range from dataset (floor to day)
            dates = np.arange(
                dataset.valid_time.dt.floor('D')[0].values,
                dataset.valid_time.dt.floor('D')[-1].values,
                np.timedelta64(1, 'D')
            )

            for day in dates:
                day = np.datetime64(day)

                # Select hours 01 to 23 of day `d`
                pr_d = dataset.pr.sel(
                    valid_time=slice(day + np.timedelta64(1, 'h'),
                                     day + np.timedelta64(23, 'h')))

                # Select 00 UTC of day `d+1` (this is hour 24 of day `d`)
                pr_d1_00 = dataset.pr.sel(
                    valid_time=day + np.timedelta64(1, 'D'))  # 00 UTC next day

                # Sum all 24 hourly values
                pr_daily = pr_d.sum(dim='valid_time') + pr_d1_00

                # Assign the time label back to day `d`
                pr_daily = pr_daily.expand_dims(valid_time=[day])

                daily_pr_list.append(pr_daily)

            # Combine into a daily DataArray
            daily_precip = xr.concat(daily_pr_list, dim='valid_time')

            # Rename the time coordinate
            daily_precip = daily_precip.rename({'valid_time': 'time'})

            # Write dataset contents to a netCDF file
            daily_precip.to_netcdf(output_file)

        # Remove original file
        os.remove(filename)

        return output_file


from setuptools import setup, find_packages

VERSION = '1.0.1' 
DESCRIPTION = 'Climxtract'
LONG_DESCRIPTION = 'ClimXtract: A Python Toolkit for Standardizing High-Resolution Climate Datasets for Regional Domains'

setup(
        name="climxtract",
        version=VERSION,
        author="Maximilian Meindl",
        author_email="<maximilian.meinndl@univie.ac.at>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        url = "https://github.com/meindlm97/ClimXtract",
        install_requires=['numpy', 'xarray', 'wget', 'cdsapi', 'cdo', 'polytope-client', 'lxml', 'conflator', 'rasterio', 'cf-units', 'esgf-pyclient', 'netcdf4', 'cfgrib', 'eccodes'],
        include_package_data=True,
        keywords=['python', 'climate', 'interpolation', 'geospatial data', 'modeling'],
        classifiers= [
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        packages=find_packages(exclude=["example_notebooks"]),
        python_requires=">=3.10",
)
